"""Create, read and write .voi files.

"""

__author__ = "Florian Krause <me@floriankrause.org>"


import os
import random
from collections import OrderedDict
from itertools import groupby

import numpy as np
import scipy.stats as stats


class VOI:
    """A class representing a VOI."""

    def __init__(self, name, data, colour=None):
        """Create a VOI object.

        Parameters
        ----------
        name : str
            the name of the VOI
        data : list
            the data of the VOI
        colour : [int, int, int], optional
            the colour of the VOI;
            a random colour will be generated if argument not given (default)

        """

        self.name = name
        self.data = np.array(data)
        if colour is not None:
            if isinstance(colour, str):
                self.colour = [int(x) for x in colour.split(" ") if x != '']
            else:
                self.colour = list(colour[:3])
        else:
            self.colour = [random.randint(0, 255),
                           random.randint(0, 255),
                           random.randint(0, 255)]

    def _format_colour(self):
        return "ColorOfVOI: " + " ".join([str(x) for x in self.colour])
    
    def _format_data(self):
        rtn = ""
        for c, coordinate in enumerate(self.data):
            rtn += " ".join([str(int(x)) for x in coordinate])
            if c < len(self.data) - 1:
                rtn += "\n"
        return rtn

    def __str__(self):
        return "NameOfVOI:  {0}\n{1}\n\nNrOfVoxels: {2}\n{3}".format(
                self.name,
                self._format_colour(),
                len(self.data),
                self._format_data())

class VOIsDefinition:
    """A class representing a VOIs definition."""

    def __init__(self, reference_space="BV", resolution=[1,1,1],
            offset=[0,0,0], framing_cube=256, left_right=1,
            subject_voi_naming="<VOI>_<SUBJ>", load=None):
        """Create a VOIs definition object.

        Parameters
        ---------- 
        reference_space : str, optional
            the reference space ("BV", "TAL", "MNI"; default="BV")
        resolution : [int, int, int], optional
            the resolution of the original VMR (default=[1,1,1])
        offset : [int, int, int], optional
            the offset of the original VMR (default=[0,0,0])
        framing_cube : int, optional
            the dimension of the original VMR framing cube (default=256)
        left_right : int, optional
            the image orientation (radiological=1, neurological=2; default=1)
        subject_voi_naming : str, optional
            the subject and VOI naming convention (default="<VOI>_<SUBJ>")
        load : str, optional
            the name of a .voi file to load

        """

        self.clear()
        self.header["OriginalVMRResolutionX"] = resolution[0]
        self.header["OriginalVMRResolutionY"] = resolution[1]
        self.header["OriginalVMRResolutionZ"] = resolution[2]
        self.header["OriginalVMROffsetX"] = offset[0]
        self.header["OriginalVMROffsetY"] = offset[1]
        self.header["OriginalVMROffsetZ"] = offset[2]
        self.header["ReferenceSpace"] = reference_space
        self.header["OriginalVMRFramingCubeDim"] = framing_cube
        self.header["LeftRightConvention"] = left_right
        self.header["SubjectVOINamingConvention"] = subject_voi_naming
        if load is not None:
            try:
                self.load(load)
            except:
                raise IOError("Could not read {0}!".format(load))

    def __str__(self):
        return "{0}\n\n{1}\n\n{2}".format(self._format_header(),
                                          self._format_vois(),
                                          self._format_vtcs())

    @property
    def header(self):
        return self._header

    @property
    def vois(self):
        return self._vois

    @property
    def vtcs(self):
        return self._vtcs

    @property
    def names(self):
        return [x.name for x in self.vois]

    @property
    def colours(self):
        return [x.colour for x in self.vois]

    def _format_header(self):
        rtn = ""
        for c,x in enumerate(self.header):
            if x in ("FileVersion", "OriginalVMRResolutionX",
                    "LeftRightConvention", "SubjectVOINamingConvention"):
                rtn += "\n"
            if x ==  "NrOfVOIs":
                rtn += "\n\n"
            rtn += "{0}:".format(x).ljust(29) + "{0}".format(self.header[x])
            if c < len(self.header) - 1:
                rtn += "\n"
            if c == 0:
                rtn += "\n"
        return rtn

    def _format_vois(self):
        rtn = "NrOfVOIs".ljust(29) + "{0}\n\n".format(len(self.vois))
        for c, voi in enumerate(self.vois):
            rtn += str(voi)
            if c < len(self.vois) - 1:
                rtn += "\n\n"
        return rtn

    def _format_vtcs(self):
        rtn = "NrOfVOIVTCs: {0}".format(len(self.vtcs))
        for c, vtc in enumerate(self.vtcs):
            rtn += vtc
            if c < len(self.vtcs) - 1:
                rtn += "\n"
        return rtn

    def add_voi(self, voi):
        """Add a VOI to the VOIs definition.

        Parameters
        ----------
        voi : VOI object
            the VOI to add to the VOIs definition

        """

        self._vois.append(voi)

    def add_vtc(self, vtc):
        """Add a VOI VTC to the VOIs definition.

        Parameters
        ----------
        vtc : str
            the full path of the VOI VTC to add to the VOIs definition

        """

        self._vtcs.append(vtc)

    def clear(self):
        """Clear VOIs definition."""

        self._header = OrderedDict([("FileVersion", 4),
                                    ("ReferenceSpace", "BV"),
                                    ("OriginalVMRResolutionX", 1),
                                    ("OriginalVMRResolutionY", 1),
                                    ("OriginalVMRResolutionZ", 1),
                                    ("OriginalVMROffsetX", 0),
                                    ("OriginalVMROffsetY", 0),
                                    ("OriginalVMROffsetZ", 0),
                                    ("OriginalVMRFramingCubeDim", 256),
                                    ("LeftRightConvention", 1),
                                    ("SubjectVOINamingConvention", "<VOI>_<SUBJ>")])
        self._vois = []
        self._vtcs = []

    def load(self, filename):
        """Load VOIs definition from a .voi file.

        This will overwrite current header and VOIs!

        Parameters
        ----------
        filename : str
            the name of the .voi file

        """

        self.clear()
        with open(filename) as f:
            lines = f.readlines()
        lines = [x[0] for x in groupby(lines)]
        for counter, line in enumerate(lines):
            if line.startswith('NrOfVOIs'):
                break
        for line in lines[:counter-1]:
            if line.strip() != "":
                value = line[27:].strip().split(" ")
                try:
                    value = int(value[0])
                except:
                    value = value[0]
                self._header[line[:27].strip(": ")] = value

        idx = counter
        for line in lines[counter+1:]: 
            idx += 1
            if line.startswith("NrOfVOIVTCs"):
                break
            if line.startswith("NameOfVOI"):
                try:
                    name = line.split(":")[1].strip()
                    colour = [int(x) for x in lines[idx+1].split(":")[1].strip().split(" ")]
                    nvoxels =  int(lines[idx+3].split(":")[1].strip())
                    start = idx + 4
                    end = start + nvoxels
                    data = []
                    for x in range(start, end):
                        data.append([int(i) for i in lines[x].split(" ") if i])
                    self.add_voi(VOI(name, data, colour))
                except IndexError:
                    pass

        for line in lines[idx+1:]:
            try:
                self.add_vtc(line.strip())
            except IndexError:
                pass

    def save(self, filename):
        """Save VOIs definition to a .voi file.

        Parameters
        ----------
        filename : str
            the name of the .voi file

        """

        if not os.path.splitext(filename)[-1].lower() == ".voi":
            filename += ".voi"

        with open(filename, 'w') as f:
            f.write(self._format_header() + "\n")
            f.write("\n\n")
            f.write(self._format_vois() + "\n")
            f.write("\n\n")
            f.write(self._format_vtcs())



"""Create, read and write .ctr files.

"""

__author__ = "Florian Krause <me@floriankrause.org>"


import os
import random
from collections import OrderedDict

import numpy as np
import scipy.stats as stats


class Contrast:
    """A class representing a Contrast object."""

    def __init__(self, name, data):
        """Create a Contrast object.

        Parameters
        ----------
        name : str
            the name of the contrast
        data : list
            the data of the contrast

        """

        self.name = name
        self.data = np.array(data)

    def __str__(self):
        return "{0}\n{1}".format(repr(self.name),
                                 repr(self.data))


class ContrastsDefinition:
    """A class representing a contrasts definition."""

    def __init__(self, load=None):
        """Create a contrasts definition object.

        Parameters
        ----------
        load : str, optional
            the name of a .ctr file to load

        """

        self.clear()
        if load is not None:
            self.load(load)

    def __str__(self):
        return "{0}\n\n{1}\n{2}".format(self._format_header(),
                                 self._format_names(),
                                 self._format_data())

    @property
    def header(self):
        return self._header

    @property
    def contrasts(self):
        return self._contrasts

    @property
    def names(self):
        return [x.name for x in self.contrasts]

    @property
    def data(self):
        for c,x in enumerate(self.contrasts):
            if c == 0:
                rtn = np.array([x.data]).T
            else:
                rtn = np.append(rtn, np.array([x.data]).T, axis=1)
        return rtn

    def _format_header(self):
        rtn = ""
        for c,x in enumerate(self.header):
            rtn += "{0}:".format(x).ljust(22) + "{0}".format(self.header[x])
            if c < len(self.header) - 1:
                rtn += "\n"
        return rtn

    def _format_names(self):
        return '"' + '" "'.join(self.names) + '"'

    def _format_data(self):
        rtn = ""
        for c,x in enumerate(self.data):
            rtn += " ".join(["{0:d}".format(y).rjust(3) for y in x])
            if c < np.shape(self.data)[0] - 1:
                rtn += "\n"
        return rtn

    def add_contrast(self, contrast):
        """Add a contrast to the contrasts definition.

        Parameters
        ----------
        contrast : Contrast object
            the contrast to add to contrasts definition

        """

        if self._header["NrOfContrasts"] == 0:
            self._header["NrOfValues"] = len(contrast.data)
        else:
            if self._header["NrOfValues"] != len(contrast.data):
                e = "Contrast has {0} data points, but contrasts definition has {1}!"
                raise ValueError(e.format(len(contrast.data),
                                          self._header["NrOfValues"]))

        self._header["NrOfContrasts"] += 1

    def clear(self):
        """Clear contrasts definition."""

        self._header = OrderedDict([("FileVersion", 1),
                                   ("NrOfContrasts", 0),
                                   ("NrOfValues", 0)])
        self._contrasts = []

    def load(self, filename):
        """Load contrasts definition from a .ctr file.

        This will overwrite current header and contrasts!

        Parameters
        ----------
        filename : str
            the name of the .ctr file

        """

        self.clear()
        with open(filename) as f:
            lines = f.readlines()
        for counter, line in enumerate(lines):
            if line.startswith('"'):
                break
        names = line.strip().strip('"').split('" "')
        for line in lines[:counter-1]:
            if line.strip():
                self._header[line[:16].strip(": ")] = int(line[16:].strip())
        data = np.array([[int(line[i:i+4]) for i in range(0, len(line), 4) if line[i].strip("\r\n")!=""]
                                       for line in lines[counter + 1:]])
        for x in range(len(names)):
            self._contrasts.append(Contrast(names[x], data[:,x]))

    def save(self, filename):
        """Save contrast definition to a .ctr file.

        Parameters
        ----------
        filename : str
            the name of the .ctr file

        """

        if not os.path.splitext(filename)[-1].lower() == ".ctr":
            filename += ".ctr"

        with open(filename, 'w') as f:
            f.write("\n")
            f.write(self._format_header() + "\n")
            f.write("\n")
            f.write(self._format_names() + "\n")
            f.write(self._format_data())



"""Create, read and write .sdm files.

"""

__author__ = "Florian Krause <me@floriankrause.org>"


import os
import random
from collections import OrderedDict

import numpy as np
import scipy.stats as stats


class Predictor:
    """A class representing a predictor."""

    def __init__(self, name, data, colour=None):
        """Create a Predictor object.

        Parameters
        ----------
        name : str
            the name of the predictor
        data : list
            the data of the predictor
        colour : [int, int, int], optional
            the colour of the regressor;
            a random colour will be generated if argument not given (default)

        """

        self.name = name
        self.data = np.array(data)
        if colour is not None:
            if isinstance(colour, str):
                self.colour = [int(x) for x in colour.split(" ") if x != '']
            else:
                self.colour = list(colour[:3])
        else:
            self.colour = [random.randint(0, 255),
                           random.randint(0, 255),
                           random.randint(0, 255)]

    def __str__(self):
        return "{0}\n{1}\n{2}\n{3}".format(self.colour, repr(self.name),
                                           repr(self.data), len(self.data))

    def _hrf(self, tr, p=[6,16,1,1,6,0,32]):
        """Create an HRF from two gamma functions.

        Paramters
        ---------
        tr : float
            repetition time (in seconds) at which to sample the HRF
        p : list, optional
            parameters of the two gamma functions:
                                                                defaults
                                                                (seconds)
            p[0] - delay of response (relative to onset)         6
            p[1] - delay of undershoot (relative to onset)      16
            p[2] - dispersion of response                        1
            p[3] - dispersion of undershoot                      1
            p[4] - ratio of response to undershoot               6
            p[5] - onset (seconds)                               0
            p[6] - length of kernel (seconds)                   32

        """

        p = [float(x) for x in p]
        tr=float(tr)
        fMRI_T = 16.0

        # HRF in seconds
        dt = tr/fMRI_T
        u = np.arange(p[6] / dt + 1) - p[5] / dt
        g1 = stats.gamma.pdf(u, p[0] / p[2], scale=1.0 / (dt / p[2]))
        g2 = stats.gamma.pdf(u, p[1] / p[3], scale=1.0 / (dt / p[3])) / p[4]
        hrf = g1 - g2

        # Sample in volumes
        good_pts = np.array(range(np.int((p[6] + tr) / tr))) * fMRI_T
        hrf = hrf[[int(x) for x in list(good_pts)]]
        hrf = hrf / np.sum(hrf);

        return hrf

    def convolve_with_hrf(self, tr):  # TODO: Needs testing!
        """Convolve with the haemodynamic response function.

        Parameters
        ----------
        tr : int
            the repetition time in milliseconds

        """

        tr = float(tr) / 1000
        hrf_ = self._hrf(tr)
        self.data = np.convolve(self.data, self._hrf(tr))[0:len(self.data)]

    def ztransform(self):
        """z-transform the predictor data."""

        self.data = stats.mstats.zscore(self.data, axis=0)

    def get_derivative(self, n=1):  # TODO: Needs testing!
        """Get derivative of predictor.

        Parameters
        ----------
        n : int
            the order of the derivative

        Returns
        -------
        predictor : Predictor
            the derivative predictor

        """

        name = self.name + "_D{0}".format(n)
        data = self.data
        for x in range(n):
            data = np.ediff1d(data, to_begin=0)

        return Predictor(name, data, self.colour)


class DesignMatrix:
    """A class representing a single-subject design matrix."""

    def __init__(self, load=None):
        """Create a design matrix object.

        Parameters
        ----------
        load : str, optional
            the name of a .sdm file to load

        """

        self.clear()
        if load is not None:
            try:
                self.load(load, col_length=11)
            except:
                try:
                    self.load(load, col_length=12)
                except:
                    raise IOError("Could not read {0}!".format(load))

    def __str__(self):
        return "{0}\n\n{1}\n{2}\n{3}".format(self._format_header(),
                                             self._format_colours(),
                                             self._format_names(),
                                             self.data)

    @property
    def header(self):
        return self._header

    @property
    def predictors(self):
        return self._predictors

    @property
    def names(self):
        return [x.name for x in self.predictors]

    @property
    def colours(self):
        return [x.colour for x in self.predictors]

    @property
    def data(self):
        for c,x in enumerate(self.predictors):
            if c == 0:
                rtn = np.array([x.data]).T
            else:
                rtn = np.append(rtn, np.array([x.data]).T, axis=1)
        return rtn

    def _format_header(self):
        rtn = ""
        for c,x in enumerate(self.header):
            rtn += "{0}:".format(x).ljust(24) + "{0}".format(self.header[x])
            if c < len(self.header) - 1:
                rtn += "\n"
            if c == 0:
                rtn += "\n"
        return rtn

    def _format_colours(self):
        colours = ["{0} {1} {2}".format(x[0],x[1],x[2]) for x in self.colours]
        return "   ".join(colours)

    def _format_names(self):
        return '"' + '" "'.join(self.names) + '"'

    def _format_data(self):
        rtn = ""
        for c,x in enumerate(self.data):
            rtn += "".join(["{0:f}".format(y).rjust(12) for y in x])
            if c < np.shape(self.data)[0] - 1:
                rtn += "\n"
        return rtn

    def add_predictor(self, predictor, is_confound=False):
        """Add a predictor to the design matrix.

        Parameters
        ----------
        predictor : Predictor object
            the predictor to add to the design matrix
        is_confound : bool, optional
            whether the predictor is a confound predictor (default=False)

        """

        if self._header["NrOfPredictors"] > 0:
            if len(predictor.data) != self._header["NrOfDataPoints"]:
                e = "Predictor has {0} data points, but design matrix has {1}!"
                raise ValueError(e.format(len(predictor.data),
                                          self._header["NrOfDataPoints"]))

        if is_confound:
            position = self._header["NrOfPredictors"] - \
                        self._header["IncludesConstant"]
        else:
            position = self._header["FirstConfoundPredictor"] - 1
        self._predictors.insert(position, predictor)

        if self._header["NrOfPredictors"] == 0:
            self._header["NrOfDataPoints"] = len(predictor.data)
        self._header["NrOfPredictors"] += 1

        if not is_confound:
            self._header["FirstConfoundPredictor"] += 1

    def add_confound_predictor(self, predictor):
        """Add a confound predictor to the design matrix.

        Parameters
        ----------
        predictor : Predictor object
            the predictor to add to the design matrix

        """

        self.add_predictor(predictor, is_confound=True)

    def add_constant(self):
        """Add constant."""

        self.add_predictor(Predictor("Constant", len(self.data) * [1], [255,255,255]), True)
        self._header["IncludesConstant"] = 1

    def define_predictors(self, protocol, data_points, tr):
        """Define predictors from stimulation protocol.

        Note: This does (for now) only create demeaned parametric predictors!

        Parameters
        ----------
        protocol : brainvoyagertools.prt.StimulationProtocol
            the protocol do define predictors from
        data_points : int
            the number of data points per predictor
        tr : int or None
            the TR for convolving with HRF; no convolution if None

        """

        if protocol.time_units != "Volumes":
            raise NotImplementedError(
                "Currently only protocols with time unit 'Volumes' supported!")

        for condition in protocol.conditions:
            # main effect
            mod = False
            parameters = []
            data = data_points * [0]
            for event in condition.data:
                if protocol.header["ParametricWeights"] == 1:
                    parameters.append(event[2])
                    if event[2] != 1:
                        mod = True
                for x in range(event[0]-1, event[1]):
                    data[x] = 1
            if mod:
                predictor = Predictor(condition.name + " [Main]", data,
                                      condition.colour)
            else:
                predictor = Predictor(condition.name, data, condition.colour)
            if tr is not None:
                predictor.convolve_with_hrf(tr)
            self.add_predictor(predictor)

            # parametric effect
            if mod:
                mean = np.mean(list(set(parameters)))
                mod_data = data_points * [0]
                for event in condition.data:
                    for x in range(event[0]-1, event[1]):
                        mod_data[x] = event[2] - mean
                predictor = Predictor(condition.name + " [Parametric]", mod_data,
                                      condition.colour)
                if tr is not None:
                    predictor.convolve_with_hrf(tr)
                self.add_predictor(predictor)

    def clear(self):
        """Clear design matrix."""

        self._header = OrderedDict([("FileVersion", 1),
                                   ("NrOfPredictors", 0),
                                   ("NrOfDataPoints", 0),
                                   ("IncludesConstant", 0),
                                   ("FirstConfoundPredictor", 1)])
        self._predictors = []

    def load(self, filename, col_length=11):
        """Load design matrix from a .sdm file.

        This will overwrite current header and predictors!

        Parameters
        ----------
        filename : str
            the name of the .sdm file
        col_length : int, optional
            the length of the columns in characters
                design files -- 11 (default)
                motion files -- 12

        """

        self.clear()
        with open(filename) as f:
            lines = f.readlines()
        for counter, line in enumerate(lines):
            if line.startswith('"'):
                break
        names = line.strip().strip('"').split('" "')
        colours = lines[counter-1].strip().split('   ')
        for line in lines[:counter-1]:
            if line.strip():
                self._header[line[:24].strip(": ")] = int(line[24:].strip())
        data = np.array([[float(line[i:i+col_length]) for i in range(0, len(line), col_length) if line[i].strip("\r\n")!=""]
                                       for line in lines[counter + 1:]])
        for x in range(len(names)):
            self._predictors.append(Predictor(names[x], data[:,x], colours[x]))

    def save(self, filename):
        """Save design matrix to a .sdm file.

        Parameters
        ----------
        filename : str
            the name of the .sdm file

        """

        if not os.path.splitext(filename)[-1].lower() == ".sdm":
            filename += ".sdm"

        with open(filename, 'w') as f:
            f.write(self._format_header() + "\n")
            f.write("\n")
            f.write(self._format_colours() + "\n")
            f.write(self._format_names() + "\n")
            f.write(self._format_data())



"""Create, read and write .mdm files.

"""

__author__ = "Florian Krause <me@floriankrause.org>"


import os
import random
from collections import OrderedDict

import numpy as np
import scipy.stats as stats


class Study:
    """A class representing a Study object."""

    def __init__(self, data_files, sdm_file):
        """Create a Study object.

        Parameters
        ----------
        data_files : str or list
            full path of .vtc file or full paths of .ssm and .mtc files
        sdm_file : str
            fullpath of .sdm file

        """

        if not isinstance(data_files, list):
            data_files = [data_files]
        self.data_files = data_files
        self.sdm_file = sdm_file

    def __str__(self):
        return "{0}\n{1}".format(self.data_files, repr(self.sdm_file))


class DesignMatrix:
    """A class representing a multi-subject design matrix."""

    def __init__(self, load=None):
        """Create a design matrix object.

        Parameters
        ----------
        load : str, optional
            the name of a .mdm file to load

        """

        self.clear()
        if load is not None:
            self.load(load)

    def __str__(self):
        return "{0}\n{1}".format(self._format_header(),
                                 self._format_studies())

    @property
    def header(self):
        return self._header

    @property
    def studies(self):
        return self._studies

    @property
    def rfx_glm(self):
        return bool(self._header["RFX-GLM"])

    @rfx_glm.setter
    def rfx_glm(self, value):
        self._header["RFX-GLM"] = int(value)

    @property
    def transformation(self):
        if self._header["PSCTransformation"] == 1:
            return "psc"
        elif self._header["zTransformation"] == 1:
            return "z"

    @transformation.setter
    def transformation(self, value):
        if value.lower() == "psc":
            self._header["PSCTransformation"] = 1
            self._header["zTransformation"] = 0
        elif value.lower() == "z":
            self._header["zTransformation"] = 1
            self._header["PSCTransformation"] = 0
        else:
            raise ValueError("'rfx_glm' can only be 'psc' or 'z'!")

    @property
    def separate_predictors(self):
        return bool(self._header["SeparatePredictors"])

    @separate_predictors.setter
    def separate_predictors(self, value):
        self._header["SeparatePredictors"] = int(value)

    def _format_header(self):
        rtn = ""
        for c,x in enumerate(self.header):
            rtn += "{0}:".format(x).ljust(22) + "{0}".format(self.header[x])
            if c < len(self.header) - 1:
                rtn += "\n"
            if c in [1,2,5]:
                rtn += "\n"
        return rtn

    def _format_studies(self):
        rtn = ""
        for c,x in enumerate(self.studies):
            rtn += '"' + '" "'.join(x.data_files) + '" "' +  x.sdm_file + '"'
            if c < self.header["NrOfStudies"] - 1:
                rtn += "\n"
        return rtn

    def add_study(self, study):
        """Add a study to the design matrix.

        Parameters
        ----------
        study : Study object
            the study to add to the design matrix

        """

        if len(study.data_files) > 1:
            if self._header["NrOfStudies"] > 0 and \
               self._header["TypeOfFunctionalData"] == "VTC":
                raise ValueError(
                    "Data is MTC, but design matrix contains VTC data!")
            else:
                self._header["TypeOfFunctionalData"] = "MTC"
        else:
            if self._header["NrOfStudies"] > 0 and \
               self._header["TypeOfFunctionalData"] == "MTC":
                raise ValueError(
                    "Data is VTC, but design matrix contains MTC data!")

        self._header["NrOfStudies"] += 1
        self._studies.append(study)

    def clear(self):
        """Clear design matrix."""

        self._header = OrderedDict([("FileVersion", 3),
                                   ("TypeOfFunctionalData", "VTC"),
                                   ("RFX-GLM", 0),
                                   ("PSCTransformation", 1),
                                   ("zTransformation", 0),
                                   ("SeparatePredictors", 0),
                                   ("NrOfStudies", 0)])
        self._studies = []

    def load(self, filename):
        """Load design matrix from a .mdm file.

        This will overwrite current header and studies!

        Parameters
        ----------
        filename : str
            the name of the .mdm file

        """

        self.clear()
        with open(filename) as f:
            lines = f.readlines()
        for counter, line in enumerate(lines):
            if line.startswith('"'):
                break
        for line in lines[:counter-1]:
            if line.strip():
                try:
                    self._header[line[:22].strip(": ")] = int(line[22:].strip())
                except ValueError:
                    self._header[line[:22].strip(": ")] = line[22:].strip()
        for line in lines[counter:]:
            tmp = [x for x in line.split('"') if x.strip()!= ""]
            if len(tmp) > 2:
                data_files = tmp[:2]
                sdm_file = tmp[2]
            else:
                data_files = [tmp[0]]
                sdm_file = tmp[1]
            self._studies.append(Study(data_files, sdm_file))

    def save(self, filename):
        """Save design matrix to a .mdm file.

        Parameters
        ----------
        filename : str
            the name of the .mdm file

        """

        if not os.path.splitext(filename)[-1].lower() == ".mdm":
            filename += ".mdm"

        with open(filename, 'w') as f:
            f.write("\n")
            f.write(self._format_header() + "\n")
            f.write(self._format_studies())



import os
import setuptools


PACKAGE_NAME = "brainvoyagertools"

def get_version():
    """Get version number."""

    with open(os.path.join("src", PACKAGE_NAME, "__init__.py")) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split("'")[1]
    return "None"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = PACKAGE_NAME,
    description=\
    'BrainVoyager Tools - ' \
    'An object-oriented approach to create, read and write common plain-text BrainVoyager input formats with Python',
    author='Florian Krause',
    author_email='me@floriankrause.org',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url = 'https://github.com/fladd/brainvoyagertools',
    version=get_version(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=['numpy>=1,<2',
                      'scipy>=1,<2',
                      ],
)



"""Create, read and write .prt files.

"""

__author__ = "Florian Krause <me@floriankrause.org>"


import os
import random
from collections import OrderedDict
from itertools import groupby

import numpy as np


class Condition:
    """A class implementing a condition."""

    def __init__(self, name, data, colour=None):
        """Create a condition.

        Parameters
        ----------
        name : str
            the name of the condition
        data : list
            the data of the condition
        colour : [int, int, int], optional
            the colour of the condition;
            a random colour will be generated if argument not given (default)

        """

        self.name = name
        self.data = np.array(data, dtype=object)
        if colour is not None:
            if isinstance(colour, str):
                self.colour = [int(x) for x in colour.split(" ") if x != '']
            else:
                self.colour = list(colour[:3])
        else:
            self.colour = [random.randint(0, 255),
                           random.randint(0, 255),
                           random.randint(0, 255)]

    def __str__(self):
        return "{0}\n{1}\n{2}\n{3}".format(self.name,
                                           len(self.data),
                                           self._format_data(),
                                           self._format_colour())

    def __add__(self, other):
        name = self.name + "+" + other.name
        data = np.sort(np.concatenate((self.data, other.data)), axis=0)
        colour = [min(self.colour[0] + other.colour[0], 255),
                 min(self.colour[1] + other.colour[1], 255),
                 min(self.colour[2] + other.colour[2], 255)]

        return Condition(name, data, colour)
    
    def _format_data(self):
        rtn = ""
        max = 0
        for intervall in self.data:
            if len(str(intervall[1])) > max:
                max = len(str(intervall[1]))
        for c, intervall in enumerate(self.data):
            rtn += " {0}".format(
                "  ".join([str(x).rjust(max) for x in intervall]))
            if c < len(self.data) - 1:
                rtn += "\n"
        return rtn

    def _format_colour(self):
        return "Color: " + " ".join([str(x) for x in self.colour])


class StimulationProtocol:
    """A class implementing a stimulation protocol."""

    def __init__(self, experiment_name="untitled", time_units="Volumes",
                 load=None):
        """Create a stimulation protocol.

        Parameters
        ----------
        experiment_name : str
            the name of the experiment (default="untitled")
        time_units : str
            "Volumes" or "msec" (default="Volumes")
        load : str, optional
            the name of a .prt file to load

        """

        self.clear()
        self._header["Experiment"] = experiment_name
        if time_units not in ["Volumes", "msec"]:
            raise ValueError("")
        self._header["ResolutionOfTime"] = time_units
        if load is not None:
            try:
                self.load(load)
            except:
                raise IOError("Could not read {0}!".format(load))

    def __str__(self):
        return "{0}\n\n{1}".format(self._format_header(),
                                   self._format_conditions())

    @property
    def header(self):
        return self._header

    @property
    def conditions(self):
        return self._conditions

    @property
    def time_units(self):
        return self._header["ResolutionOfTime"]

    @property
    def experiment_name(self):
        return self._header["Experiment"]

    @property
    def event_list(self):
        events = []
        for condition in self.conditions:
            for intervall in condition.data:
                event = [condition.name]
                event.extend(intervall)
                events.append(event)
        return sorted(events, key=lambda x: x[1])

    @property
    def event_names(self):
        return [x[0] for x in self.event_list]

    @property
    def event_onsets(self):
        return [x[1] for x in self.event_list]

    @property
    def event_durations(self):
        if self.time_units == "msec":
            return [x[2] - x[1] for x in self.event_list]
        else:
            return [x[2] - x[1] + 1 for x in self.event_list]

    @property
    def event_weights(self):
        try:
            return [x[3] for x in self.event_list]
        except IndexError:
            return [1 for x in self.event_list]

    @property
    def condition_names(self):
        return [c.name for c in self.conditions]

    @property
    def condition_onsets(self):
        return [c.data[:,0] for c in self.conditions]

    @property
    def condition_durations(self):
        if self.time_units == "msec":
            return [c.data[:,1] - c.data[:,0] for c in self.conditions]
        else:
            return [c.data[:,1] - c.data[:,0] + 1 for c in self.conditions]

    @property
    def condition_weights(self):
        try:
            return [c.data[:,2] for c in self.conditions]
        except IndexError:
            return [np.array([1] * len(c.data)) for c in self.conditions]

    def _format_header(self):
        rtn = ""
        for c,x in enumerate(self.header):
            if c != 9 or (c == 9 and self._header["FileVersion"] == 3):
                if isinstance(self.header[x], list):
                    r,g,b = self.header[x]
                    value = "{0} {1} {2}".format(r,g,b)
                else:
                    value = self.header[x]
                rtn += "{0}:".format(x).ljust(20) + \
                    "{0}".format(value)
                if c < len(self.header) - 1:
                    rtn += "\n"
                if c in [0, 1, 2, 8, 9]:
                    rtn += "\n"
        return rtn

    def _format_conditions(self):
        rtn = ""
        for c, condition in enumerate(self.conditions):
            rtn += str(condition)
            if c < len(self.conditions) - 1:
                rtn += "\n\n"
        return rtn

    def add_condition(self, condition):
        """Add a condition to the stimulation protocol.

        Parameters
        ----------
        condition : Condition object
            the condition to add to the stimulation protocol

        """

        if condition.data.shape[1] > 2:
            self._header["FileVersion"] = 3
            self._header["ParametricWeights"] = 1
            for condition_ in self.conditions:
                if condition_.data.shape[1] < 3:
                    condition_.data = np.append(
                        condition_.data,
                        np.array([(condition_.data.shape[0]) * [1]]).T,
                        axis=1)
        else:
            if self._header["FileVersion"] == 3 and \
               self._header["ParametricWeights"] == 1:
                condition.data = np.append(
                    condition.data,
                    np.array([(condition.data.shape[0]) * [1]]).T,
                    axis=1)

        self._conditions.append(condition)
        self._header["NrOfConditions"] += 1

    def clear(self):
        """Clear Protocol."""
        self._header = OrderedDict([("FileVersion", 2),
                                   ("ResolutionOfTime", "Volumes"),
                                   ("Experiment", "untitled"),
                                   ("BackgroundColor", [0, 0, 0]),
                                   ("TextColor", [255, 255, 255]),
                                   ("TimeCourseColor", [255, 255, 255]),
                                   ("TimeCourseThick", 3),
                                   ("ReferenceFuncColor", [0, 0, 80]),
                                   ("ReferenceFuncThick", 3),
                                   ("ParametricWeights", 0),
                                   ("NrOfConditions", 0)])

        self._conditions = []

    def convert_to_msec(self, tr):
        """Convert time units to milliseconds.

        Parameters
        ----------
        tr : int
            the repetition time in milliseconds

        """

        if self._header["ResolutionOfTime"] == "msec":
            pass
        else:
            for condition in self._conditions:
                for intervall in condition.data:
                    intervall[0] = int((intervall[0] - 1) * tr)
                    intervall[1] = int(intervall[1] * tr)
            self._header["ResolutionOfTime"] = "msec"

    def load(self, filename):
        """Load stimulation protocol from a .prt file.

        This will overwrite current header and conditions!

        Parameters
        ----------
        filename : str
            the name of the .prt file

        """

        with open(filename) as f:
            lines = f.readlines()
        lines = [x[0].replace("\t", "  ").strip() for x in groupby(lines)]
        for counter, line in enumerate(lines):
            if line.startswith('NrOfConditions'):
                break
        for line in lines[:counter]:
            if line.strip() != "":
                value = line.split(":", 1)[1].strip().split(" ")
                if len(value) > 1:
                    try:
                        value = [int(x) for x in value]
                    except ValueError:
                        value = " ".join(value)
                else:
                    try:
                        value = int(value[0])
                    except ValueError:
                        value = value[0]
                self._header[line.split(":", 1)[0].strip()] = value

        def int_or_float(x):
            try:
                return int(x)
            except:
                return float(x)

        idx = counter
        for line in lines[counter+1:]:
            idx += 1
            if not line.strip():
                try:
                    name = lines[idx+1].strip()
                    start = idx + 3
                    end = start + int(lines[idx+2])
                    data = []
                    for x in range(start, end):
                        data.append([int_or_float(i) for i in lines[x].split(" ") if i])
                    colour = [int(x) for x in lines[end][7:].split(" ")]
                    self.add_condition(Condition(name, data, colour))
                except IndexError:
                    pass

    def save(self, filename):
        """Save stimulation protocol to a .prt file.

        Parameters
        ----------
        filename : str
            the name of the .prt file

        """

        if not os.path.splitext(filename)[-1].lower() == ".prt":
            filename += ".prt"

        with open(filename, 'w') as f:
            f.write("\n")
            f.write(self._format_header() + "\n")
            f.write("\n")
            f.write(self._format_conditions())
